/*
 * Balanced RQ-based Hotplug driver for MSM SoC
 *
 * Based on drivers/cpuquiet/governors/balanced.c
 * Contains portions of code from EternityProject T3 CPUManager
 * Contains portions of code from NVIDIA CORPORATION
 *
 * Authored by: AngeloGioacchino Del Regno <kholk11@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

/* Note: NEVER suspend the runqueue averaging worker! */

#include <linux/kernel.h>
#include <linux/cpufreq.h>
#include <linux/cpuquiet.h>
#include <linux/cpumask.h>
#include <linux/module.h>
#include <linux/pm_qos.h>
#include <linux/sysfs.h>
#include <linux/suspend.h>
#include <linux/tick.h>
#include <linux/jiffies.h>
#include <linux/cpu.h>
#include <linux/sched.h>
#include <linux/slab.h>

#include "../cpuquiet.h"

#define CPQ_CUSTOM_ATTRIBUTE(_name, _mode, _show, _store)		\
	static struct cpuquiet_attribute _name ## _attr = {		\
		.attr = {.name = __stringify(_name), .mode = _mode },	\
		.show = _show,						\
		.store = _store,					\
		.param = &_name,					\
	}

#define RQB_ATTRIBUTE(_name, _mode, _show, _store)			\
	static struct cpuquiet_attribute _name ## _attr = {		\
		.attr = {.name = __stringify(_name), .mode = _mode },	\
		.show = _show,						\
		.store = _store,					\
	}

#define CPUNAMELEN 8

#define RQ_SAMPLE_TIME_NS	250000000
#define RQ_AVG_TIMER_RATE_NS	10000000
#define RQ_SAMPLE_CAPACITY	(RQ_SAMPLE_TIME_NS / RQ_AVG_TIMER_RATE_NS)

/* Don't put any CPU down prior at least 3x scheduling samplings */
#define DNCORE_DELAY_MS		(3 * (RQ_SAMPLE_TIME_NS / NSEC_PER_MSEC))
#define UPCORE_DELAY_MS		((RQ_SAMPLE_TIME_NS / NSEC_PER_MSEC) / 2)
#define RECHECK_DELAY_MS	(6 * (RQ_SAMPLE_TIME_NS / NSEC_PER_MSEC))

typedef enum {
	/* Brings one more CPU core on-line */
	CPU_UPCORE,

	/* Keeps CPU core composition unchanged */
	CPU_FAIR,

	/* Offline one CPU core */
	CPU_DNCORE,
} CPU_SPEED_BALANCE;

typedef enum {
	IDLE,
	DOWN,
	UP,
} RQBALANCE_STATE;

#define CLUSTER_LITTLE		0
#define CLUSTER_BIG		1
#define MAX_CLUSTERS		2

struct idle_info {
	u64 idle_last;
	u64 last_timestamp;
	u64 idle_current;
	u64 timestamp;
};

struct cpu_limits {
	unsigned int main_cpu;
	unsigned int freq_min;
	unsigned int freq_max;
	bool cpuhp_scheduled;
};

struct cpuquiet_governor rqbalance_governor;

static DEFINE_PER_CPU(struct idle_info, idleinfo);
static DEFINE_PER_CPU(unsigned int, cpu_load);

static struct timer_list load_timer;
static bool load_timer_active;
static bool soc_is_hmp;
static unsigned int available_clusters = 0;

/* configurable parameters */
static signed int    balance_penalty = 0; /* Range: -99, +inf */
static unsigned int  balance_level = 60;  /* Range:   1, +inf */
static unsigned int  idle_bottom_freq[MAX_CLUSTERS];
static unsigned int  idle_top_freq[MAX_CLUSTERS];
static unsigned int  num_of_cores[MAX_CLUSTERS];
static unsigned long up_delay;
static unsigned long down_delay;
static unsigned long recheck_delay;
static unsigned long last_change_time;
static unsigned int  load_sample_rate = 20; /* msec */
static struct cpu_limits rqb_freq_limits[MAX_CLUSTERS];
static enum cpuhp_state rqb_hp_online;
static struct workqueue_struct *rqbalance_wq;
static struct delayed_work rqbalance_work;
static RQBALANCE_STATE rqbalance_state;
static struct notifier_block pm_notifier_block;

/* Either online or unisolated CPUs */
const struct cpumask *avail_cpus_mask;

static void calculate_load_timer(unsigned long data)
{
	int i;
	u64 idle_time, elapsed_time;

	if (!load_timer_active)
		return;

	for_each_cpu(i, avail_cpus_mask) {
		struct idle_info *iinfo = &per_cpu(idleinfo, i);
		unsigned int *load = &per_cpu(cpu_load, i);

		iinfo->idle_last = iinfo->idle_current;
		iinfo->last_timestamp = iinfo->timestamp;
		iinfo->idle_current =
			get_cpu_idle_time_us(i, &iinfo->timestamp);
		elapsed_time = iinfo->timestamp - iinfo->last_timestamp;

		idle_time = iinfo->idle_current - iinfo->idle_last;
		idle_time *= 100;
		do_div(idle_time, elapsed_time);
		*load = 100 - idle_time;
	}
	mod_timer(&load_timer, jiffies + msecs_to_jiffies(load_sample_rate));
}

static void start_load_timer(void)
{
	int i;

	if (load_timer_active)
		return;

	load_timer_active = true;

	for_each_cpu(i, avail_cpus_mask) {
		struct idle_info *iinfo = &per_cpu(idleinfo, i);

		iinfo->idle_current =
			get_cpu_idle_time_us(i, &iinfo->timestamp);
	}
	mod_timer(&load_timer, jiffies + msecs_to_jiffies(100));
}

static void stop_load_timer(void)
{
	if (!load_timer_active)
		return;

	load_timer_active = false;
	del_timer(&load_timer);
}

static unsigned int get_slowest_cpu_n(void)
{
	unsigned int cpu = nr_cpu_ids;
	unsigned long minload = ULONG_MAX;
	int i;

	for_each_cpu(i, avail_cpus_mask) {
		unsigned int *load = &per_cpu(cpu_load, i);

		if ((i > 0) && (minload > *load)) {
			cpu = i;
			minload = *load;
		}
	}

	return cpu;
}

static unsigned int cpu_highest_speed(void)
{
	unsigned int maxload = 0;
	int i;

	for_each_cpu(i, avail_cpus_mask) {
		unsigned int *load = &per_cpu(cpu_load, i);

		maxload = max(maxload, *load);
	}

	return maxload;
}

static unsigned int count_slow_cpus(unsigned int limit)
{
	unsigned int cnt = 0;
	int i;

	for_each_cpu(i, avail_cpus_mask) {
		unsigned int *load = &per_cpu(cpu_load, i);

		if (*load <= limit)
			cnt++;
	}

	return cnt;
}

static unsigned int num_online_cluster_cpus(bool cluster)
{
	unsigned int cnt = 0;
	int i;

	for_each_cpu(i, avail_cpus_mask) {
		if (topology_physical_package_id(i) == cluster)
			cnt++;
	}

	return cnt;
}

static inline int get_offline_core(int cluster)
{
	int i;

	for_each_possible_cpu(i) {
		if (topology_physical_package_id(i) != cluster)
			continue;

		if (cpu_is_offline(i))
			return i;
	};

	return 0;
}

/*
 * There are two ladders here for hysteresis. We do not want to offline
 * and online a CPU repeatedly. Thus, the down thresholds are *lower*
 * than the corresponding up threshold.
 * The size of both arrays shall be the same.
 *
 * NOTE: On HMP multi-cluster systems, the last value is used to enable
 *       onlining of the high performance cluster CPUs, but ONLY when
 *       clusterswitch (one cluster at a time) mode is enabled.
 */
static unsigned int nr_run_thresholds[] = {
/*      1,	2,	3,	4 (target on-line num of cpus) */
	150,	400,	500,	520,

/*	5,	6,	7,	8 (HMP target on-line num of CPUs) */
	550,	640,	750,	UINT_MAX
};

static unsigned int nr_down_run_thresholds[] = {
/*	1,	2,	3,	4 (target off-line num of cpus) */
	0,	80,	290,	360,

/*	5,	6,	7,	8 (HMP target off-line num of CPUs) */
	440,	500,	550,	680
};

/*
 * Those are hysteresis values for the high performance CPUs.
 *
 * The logic in this is to add the hysteresis value as it is if we want
 * to calculate our ladders for onlining a core (upcore), but divide by
 * half (each corresponding value) if we want to calculate the ladders
 * for offlining a core (dncore).
 *
 * This logic pretends we want less power consumption at a little
 * performance cost.
 * On mobile devices with passive cooling, this is expected to increase
 * CPU clusters efficiency by reducing heat and subsequently throttling.
 *
 * NOTE: This is used to compose *complete* run thresholds arrays for
 *       the high performance CPUs cluster when clusterswitch mode
 *       is disabled and when we can use both clusters simultaneously.
 */
//static unsigned int hiperf_cpu_hysteresis[] = {
/*	1,	2,	3,	4 (used only for downcore hysteresis) */
//	15,     25,	30,	60
//};

struct runqueue_sample {
	int64_t sample_time;
	unsigned long nr_run;
};

struct runqueue_data {
	unsigned int update_rate;

	/* Circular buffer. */
	struct runqueue_sample nr_run_samples[RQ_SAMPLE_CAPACITY];

	/* Index of the head of the circular buffer = the newest element. */
	uint8_t nr_run_sample_head;

	/* Index of the tail of the circular buffer = oldest element. */
	uint8_t nr_run_sample_tail;

	/* Number of samples in the buffer. */
	uint8_t nr_run_sample_count;

	struct delayed_work work;
	struct workqueue_struct *nr_run_wq;
	spinlock_t lock;
};

static struct runqueue_data *rq_data;

static void rq_work_fn(struct work_struct *work)
{
	unsigned long flags = 0;
	int64_t cur_time = ktime_to_ns(ktime_get());
	struct runqueue_sample* sample = NULL;

	spin_lock_irqsave(&rq_data->lock, flags);
	if (unlikely(rq_data->nr_run_sample_count != RQ_SAMPLE_CAPACITY)) {
		/* Buffer still growing. */
		rq_data->nr_run_sample_head = rq_data->nr_run_sample_count;
		rq_data->nr_run_sample_tail = 0;
		++rq_data->nr_run_sample_count;
	} else {
		/* Buffer already full. We will be clobbering old samples. */
		rq_data->nr_run_sample_head =
			rq_data->nr_run_sample_head == RQ_SAMPLE_CAPACITY - 1 ?
				0 :
				rq_data->nr_run_sample_head + 1;
		rq_data->nr_run_sample_tail =
			rq_data->nr_run_sample_tail == RQ_SAMPLE_CAPACITY - 1 ?
				0 :
				rq_data->nr_run_sample_tail + 1;
	}
	sample = rq_data->nr_run_samples + (rq_data->nr_run_sample_head);

	/* Store the sample. */
	sample->sample_time = cur_time;
	sample->nr_run = nr_running();

	if (likely(rq_data->update_rate != 0))
		queue_delayed_work(rq_data->nr_run_wq, &rq_data->work,
				   msecs_to_jiffies(rq_data->update_rate));

	spin_unlock_irqrestore(&rq_data->lock, flags);
}

static unsigned int get_nr_run_avg(void)
{
	struct runqueue_sample nr_run_samples[RQ_SAMPLE_CAPACITY];
	uint8_t nr_run_sample_head;
	uint8_t nr_run_sample_count;

	size_t j;
	struct runqueue_sample* i;
	unsigned int nr_run = 0;
	unsigned long flags = 0;
	int64_t after_time = ktime_to_ns(ktime_get()) - RQ_SAMPLE_TIME_NS;

	spin_lock_irqsave(&rq_data->lock, flags);
	memcpy(nr_run_samples, rq_data->nr_run_samples, sizeof(nr_run_samples));
	nr_run_sample_head = rq_data->nr_run_sample_head;
	nr_run_sample_count = rq_data->nr_run_sample_count;
	spin_unlock_irqrestore(&rq_data->lock, flags);

	for (i = nr_run_samples + nr_run_sample_head, j = 0;
	     j < nr_run_sample_count; ++j) {
		if (i->sample_time < after_time) {
			/* This sample is older than the ones we find relevant. */
			break;
		}

		nr_run += i->nr_run;

		/* Get the previous element in the circular buffer. */
		if (unlikely(i == nr_run_samples)) {
			/* Wrap around. */
			i = nr_run_samples + (RQ_SAMPLE_CAPACITY - 1);
		} else {
			--i;
		}
	}

	if (unlikely(j == 0))
		return 0;

	return nr_run * 100 / j;
}

static void start_rq_work(void)
{
	if (rq_data->nr_run_wq == NULL)
		rq_data->nr_run_wq =
			create_freezable_workqueue("nr_run_avg");

	queue_delayed_work(rq_data->nr_run_wq, &rq_data->work,
			   msecs_to_jiffies(rq_data->update_rate));
	return;
}

static void stop_rq_work(void)
{
	if (rq_data->nr_run_wq)
		cancel_delayed_work(&rq_data->work);
	return;
}

/*
 * This is the brain of the hotplug implementation.
 *
 * This function is called by rqbalance_work_func to determine if
 * the CPU composition needs to be changed.
 */
static CPU_SPEED_BALANCE balanced_speed_balance(void)
{
	unsigned long highest_speed = cpu_highest_speed();
	unsigned long balanced_speed = 
		highest_speed * balance_level / (100 + balance_penalty);
	unsigned long skewed_speed = balanced_speed / 2;
	unsigned int nr_cpus = num_online_cpus();
	unsigned int max_cpus = pm_qos_request(PM_QOS_MAX_ONLINE_CPUS) ? : CONFIG_NR_CPUS;

	unsigned int avg_nr_run = get_nr_run_avg();
	unsigned int nr_run;

	/* First use the up thresholds to see if we need to bring CPUs online. */
	pr_debug("%s: Current core count max runqueue: %d\n", __func__,
		 nr_run_thresholds[nr_cpus - 1]);
	for (nr_run = nr_cpus;
		nr_run < ARRAY_SIZE(nr_run_thresholds);
		++nr_run) {
		if (avg_nr_run <= nr_run_thresholds[nr_run - 1])
			break;
	}

	/*
	 * Then use the down thresholds to see if we can offline CPUs.
	 * This looks counter-intuitive, but it does work. The first loop
	 * handles the case where there are more things running than our up
	 * thresholds. If we have now too many CPUs online for too few tasks,
	 * the loop above would be a no-op, and we will fall through to this
	 * down threshold comparison loop.
	 */
	for ( ; nr_run > 1; --nr_run) {
		if (avg_nr_run >= nr_down_run_thresholds[nr_run - 1]) {
			/* We have fewer things running than our down threshold.
			   Use one less CPU. */
			break;
		}
	}

	/* Consider when to take a CPU offline.  We need to have at least
	 * one extra idle CPU, or either of: */
	if (count_slow_cpus(skewed_speed) >= 2 ||

	    /* More CPUs than policy allows */
	    (nr_cpus > max_cpus) ||

	    /* Runqueue has been emptied */
	    (nr_run < nr_cpus))
		return CPU_DNCORE;

	/* Consider when we may want to keep CPU composition unchanged.
	 * We need to have no extra idle CPU, or: */
	if (count_slow_cpus(balanced_speed) >= 1 ||

	    /* Already at maximum CPUs allowed by policy */
	    (nr_cpus == max_cpus) ||

	    /* Enough CPUs to process the runqueue */
	    (nr_run <= nr_cpus))
		return CPU_FAIR;

	/* Otherwise, we have to online one more CPU */
	return CPU_UPCORE;
}

static void hotplug_smp_final_decision(unsigned int cpu, bool up)
{
	if (up)
		/* TODO: If we just put another CPU online, reduce the frequency of
		 * the CPU since we have increased the overall computational
		 * power of the processor. */
		cpuquiet_wake_cpu(cpu, false);
	else
		cpuquiet_quiesce_cpu(cpu, false);
}

static void hotplug_hmp_final_decision(unsigned int cpu, bool up)
{
	unsigned int cluster = topology_physical_package_id(cpu);
	unsigned int little_cores_on, big_cores_on;

	if (up) {
		/* Analyze online cores situation */
		big_cores_on = num_online_cluster_cpus(CLUSTER_BIG);
		little_cores_on = num_online_cluster_cpus(CLUSTER_LITTLE);

		/* If we want 4 little cores, we can instead use one
		   big core and two little */
		if (cluster == CLUSTER_LITTLE &&
		    little_cores_on >= num_of_cores[CLUSTER_LITTLE] - 1 &&
		    big_cores_on == 0) {
			cpuquiet_wake_cpu(
				get_offline_core(CLUSTER_BIG), false);
			cpuquiet_quiesce_cpu(
				get_offline_core(CLUSTER_LITTLE), false);
			return;
		};
		cpuquiet_wake_cpu(cpu, false);
	} else {
		cpuquiet_quiesce_cpu(cpu, false);
	}
}

/*
 * This is the main machinery for the hotplug implementation.
 *
 * This function will call itself at the end of approximately every up_delay
 * jiffies.
 *
 * The decision to power up or power down a CPU is made by
 * balanced_speed_balance.
 */
static void rqbalance_work_func(struct work_struct *work)
{
	bool up = false;
	unsigned int cpu = nr_cpu_ids;
	unsigned long now = jiffies;

	CPU_SPEED_BALANCE balance;

	switch (rqbalance_state) {
	case IDLE:
		break;
	case DOWN:
		cpu = get_slowest_cpu_n();
		if (cpu < nr_cpu_ids) {
			up = false;
			if (cpu_highest_speed() > 98) {
				rqbalance_state = UP;
				queue_delayed_work(rqbalance_wq,
					&rqbalance_work, recheck_delay);
				break;
			}
		} else {
			stop_load_timer();
		}

		queue_delayed_work(rqbalance_wq, &rqbalance_work, up_delay);
		break;
	case UP:
		balance = balanced_speed_balance();
		switch (balance) {
		/* cpu speed is up and balanced - one more on-line */
		case CPU_UPCORE:
			cpu = cpumask_next_zero(0, avail_cpus_mask);
			if (cpu < nr_cpu_ids)
				up = true;
			break;
		/* cpu speed is up, but skewed - remove one core */
		case CPU_DNCORE:
			cpu = get_slowest_cpu_n();
			if (cpu < nr_cpu_ids)
				up = false;
			break;
		/* cpu speed is up, but under-utilized - do nothing */
		case CPU_FAIR:
		default:
			break;
		}
		queue_delayed_work(
			rqbalance_wq, &rqbalance_work, up_delay);
		break;
	default:
		pr_err("%s: invalid cpuquiet balanced governor state %d\n",
			__func__, rqbalance_state);
	}

	if (!up && ((now - last_change_time) < down_delay))
		cpu = nr_cpu_ids;

	if (cpu < nr_cpu_ids) {
		last_change_time = now;
		if (!soc_is_hmp)
			hotplug_smp_final_decision(cpu, up);
		else
			hotplug_hmp_final_decision(cpu, up);
	}
}

static int balanced_cpufreq_transition(struct notifier_block *nb,
	unsigned long state, void *data)
{
	struct cpufreq_freqs *freqs = data;
	unsigned long cpu_freq;
	int n = 0;

	/*
	 * If we have at least one BIG CPU online, then take
	 * a decision based on calculations made on BIG.
	 */
	if (soc_is_hmp)
		if (num_online_cluster_cpus(CLUSTER_BIG) > 0)
			n = 1;

	if (state == CPUFREQ_POSTCHANGE) {
		cpu_freq = freqs->new;

		switch (rqbalance_state) {
		case IDLE:
			if (cpu_freq >= idle_top_freq[n]) {
				rqbalance_state = UP;
				queue_delayed_work(
					rqbalance_wq, &rqbalance_work, up_delay);
				start_load_timer();
			} else if (cpu_freq <= idle_bottom_freq[n]) {
				rqbalance_state = DOWN;
				queue_delayed_work(
					rqbalance_wq, &rqbalance_work,
					down_delay);
				start_load_timer();
			}
			break;
		case DOWN:
			if (cpu_freq >= idle_top_freq[n]) {
				rqbalance_state = UP;
				queue_delayed_work(
					rqbalance_wq, &rqbalance_work, up_delay);
				start_load_timer();
			}
			break;
		case UP:
			if (cpu_freq <= idle_bottom_freq[n])
				rqbalance_state = DOWN;

			queue_delayed_work(rqbalance_wq,
				&rqbalance_work, up_delay);
			start_load_timer();
			break;
		default:
			pr_err("%s: invalid cpuquiet rqbalance governor "
				"state %d\n", __func__, rqbalance_state);
		}
	}

	return NOTIFY_OK;
}

static struct notifier_block balanced_cpufreq_nb = {
	.notifier_call = balanced_cpufreq_transition,
};

static int frequency_limits_set(struct notifier_block *nb,
	unsigned long event, void *pol)
{
	struct cpufreq_policy *cfpol = pol;
	unsigned int cluster;

	if (event != CPUFREQ_ADJUST)
		return NOTIFY_OK;

	cluster = topology_physical_package_id(cfpol->cpu);

	cpufreq_verify_within_limits(cfpol,
					rqb_freq_limits[cluster].freq_min,
					rqb_freq_limits[cluster].freq_max);

	return NOTIFY_OK;
}

static struct notifier_block frequency_limits_nb = {
	.notifier_call = frequency_limits_set,
};

static int cfl_hotplug_notify(unsigned int cpu)
{
	unsigned int cluster;

	cluster = topology_physical_package_id(cpu);
	if (!rqb_freq_limits[cluster].cpuhp_scheduled)
		return 0;

	rqb_freq_limits[cluster].cpuhp_scheduled = false;
	cpufreq_update_policy(rqb_freq_limits[cluster].main_cpu);

	return 0;
}

static ssize_t store_ulong_delay(struct cpuquiet_attribute *cattr, const char *buf,
				size_t count)
{
	int err;
	unsigned long val;

	err = kstrtoul(buf, 0, &val);
	if (err < 0)
		return err;

	*((unsigned long *)(cattr->param)) = msecs_to_jiffies(val);

	return count;
}

static ssize_t show_ulong_delay(struct cpuquiet_attribute *cattr, char *buf)
{
	return sprintf(buf, "%u\n", jiffies_to_msecs(*((unsigned long *)cattr->param)));
}

static inline size_t array_size_from_name(const char* name)
{
	if (!strncmp(name, "idle_bottom_freq", strlen(name)-2))
		return ARRAY_SIZE(idle_bottom_freq);

	if (!strncmp(name, "idle_top_freq", strlen(name)-2))
		return ARRAY_SIZE(idle_top_freq);

	if (!strncmp(name, "nr_down_run_thresholds", strlen(name)-2))
		return ARRAY_SIZE(nr_down_run_thresholds);

	if (!strncmp(name, "nr_run_thresholds", strlen(name)-2))
		return ARRAY_SIZE(nr_run_thresholds);

	return -EINVAL;
}

static ssize_t store_uint_array(struct cpuquiet_attribute *cattr,
				const char *buf, size_t count)
{
	const char *i;
	uint8_t j;
	size_t sz = array_size_from_name(cattr->attr.name);

	for (i = buf, j = 0; *i && (j < sz); j++) {
		int read = 0;
		unsigned int val = 0;

		sscanf(i, "%u%n", &val, &read);
		if (!read)
			break;

		i += read;
		((unsigned int*)cattr->param)[j] = val;
	}

	/* If we are setting nr_run_thresholds ensure
	   the target MAX is set */
	if (!strncmp(cattr->attr.name, "nr_run_thresholds",
				strlen(cattr->attr.name)-2)) {
		int x, max_cpu_id = 0;

		for_each_possible_cpu(x)
			max_cpu_id++;

		/* Set value as target MAX on-line number of CPUs */
		if (((unsigned int*)cattr->param)[max_cpu_id] != UINT_MAX &&
			max_cpu_id < sz) {
			((unsigned int*)cattr->param)[max_cpu_id] = UINT_MAX;
		}
	}

	return count;
}

static ssize_t show_uint_array(struct cpuquiet_attribute *cattr,
					char *buf)
{
	int i;
	char *temp = buf;
	size_t sz = array_size_from_name(cattr->attr.name);

	for (i = 0; i < sz; i++)
		temp += sprintf(temp, "%u ", ((unsigned int*)cattr->param)[i]);

	temp += sprintf(temp, "\n");
	return temp - buf;
}

static inline void __set_cluster_vote(unsigned int cluster)
{
	struct cpufreq_policy cfpol;

	if (cpufreq_get_policy(&cfpol, rqb_freq_limits[cluster].main_cpu)) {
		/* The cluster is down. Schedule for the next upcore event */
		rqb_freq_limits[cluster].cpuhp_scheduled = true;

		/* Make sure we upcore ASAP to set the limits */
		cpuquiet_wake_cpu(rqb_freq_limits[cluster].main_cpu, false);
	} else {
		/* The cluster is available. Set limits right now! */
		get_online_cpus();
		cpufreq_update_policy(rqb_freq_limits[cluster].main_cpu);
		put_online_cpus();
	}
}

/*
 * Set a vote for MAX cluster frequency.
 * The function wants a string that contains "NCLUSTER FREQUENCY(KHz)".
 * To remove the vote, the string shall contain "NCLUSTER 0" or
 * "NCLUSTER UINT_MAX" (where UINT_MAX shall be a number).
 * 
 * For example, to set a limit of 133MHz on CLUSTER_LITTLE:
 * "0 133000", where 0 is CLUSTER_LITTLE and 133000 is frequency is KHz;
 * to remove this vote, the string would be "0 0".
 *
 * Eventual thermal driver votes are respected because the function
 * cpufreq_verify_within_limits will automatically take care of this.
 */
static ssize_t set_cluster_vote_max(struct cpuquiet_attribute *cattr,
					const char *buf, size_t count)
{
	unsigned int cluster, req_freq;

	if (sscanf(buf, "%u %u", &cluster, &req_freq) != 2)
		return -EINVAL;

	if (req_freq == 0)
		req_freq = UINT_MAX;

	rqb_freq_limits[cluster].freq_max = req_freq;

	__set_cluster_vote(cluster);

	return count;
}

static ssize_t set_cluster_vote_min(struct cpuquiet_attribute *cattr,
					const char *buf, size_t count)
{
	unsigned int cluster, req_freq;

	if (sscanf(buf, "%u %u", &cluster, &req_freq) != 2)
		return -EINVAL;

	rqb_freq_limits[cluster].freq_min = req_freq;

	__set_cluster_vote(cluster);

	return count;
}

/*
 * Get the current cluster votes in a human readable format.
 *
 * This function prints the status of the current votes.
 * In case there is no MAX vote, the printed value will be 0,
 * to avoid printing a very large number with no meaning.
 */
static ssize_t get_cluster_votes(struct cpuquiet_attribute *cattr,
					char *buf)
{
	return sprintf(buf, "CLUSTER   %9s%9s\n"
			    "LITTLE    %9u%9u\n"
			    "BIG       %9u%9u\n",
		"MIN", "MAX",
		rqb_freq_limits[CLUSTER_LITTLE].freq_min,
		(rqb_freq_limits[CLUSTER_LITTLE].freq_max == UINT_MAX ?
		 0 : rqb_freq_limits[CLUSTER_LITTLE].freq_max),
		rqb_freq_limits[CLUSTER_BIG].freq_min,
		(rqb_freq_limits[CLUSTER_BIG].freq_max == UINT_MAX ?
		 0 : rqb_freq_limits[CLUSTER_BIG].freq_max));
}

CPQ_SIMPLE_ATTRIBUTE(balance_level, 0644, uint);
CPQ_SIMPLE_ATTRIBUTE(balance_penalty, 0644, int);
CPQ_SIMPLE_ATTRIBUTE(load_sample_rate, 0644, uint);
CPQ_CUSTOM_ATTRIBUTE(up_delay, 0644,
			show_ulong_delay, store_ulong_delay);
CPQ_CUSTOM_ATTRIBUTE(down_delay, 0644,
			show_ulong_delay, store_ulong_delay);
CPQ_CUSTOM_ATTRIBUTE(idle_bottom_freq, 0644,
			show_uint_array, store_uint_array);
CPQ_CUSTOM_ATTRIBUTE(idle_top_freq, 0644,
			show_uint_array, store_uint_array);
CPQ_CUSTOM_ATTRIBUTE(nr_down_run_thresholds, 0644,
			show_uint_array, store_uint_array);
CPQ_CUSTOM_ATTRIBUTE(nr_run_thresholds, 0644,
			show_uint_array, store_uint_array);
RQB_ATTRIBUTE(cluster_freq_vote_max, 0644,
			get_cluster_votes, set_cluster_vote_max);
RQB_ATTRIBUTE(cluster_freq_vote_min, 0644,
			get_cluster_votes, set_cluster_vote_min);

static struct attribute *rqbalance_attrs[] = {
	&balance_level_attr.attr,
	&balance_penalty_attr.attr,
	&load_sample_rate_attr.attr,
	&up_delay_attr.attr,
	&down_delay_attr.attr,
	&idle_bottom_freq_attr.attr,
	&idle_top_freq_attr.attr,
	&nr_down_run_thresholds_attr.attr,
	&nr_run_thresholds_attr.attr,
	&cluster_freq_vote_max_attr.attr,
	&cluster_freq_vote_min_attr.attr,
	NULL,
};

static struct attribute_group rqbalance_group = {
	.name = "rqbalance",
	.attrs = rqbalance_attrs,
};

static int rqbalance_sysfs_init(void)
{
	return cpuquiet_register_attrs(&rqbalance_group);
}

static void rqbalance_sysfs_exit(void)
{
	cpuquiet_unregister_attrs(&rqbalance_group);
}

static void rqbalance_kickstart(void)
{
	struct cpufreq_freqs initial_freq;

	/*FIXME: Kick start the state machine by faking a freq notification*/
	initial_freq.new = cpufreq_get(0);
	if (initial_freq.new != 0)
		balanced_cpufreq_transition(NULL, CPUFREQ_POSTCHANGE,
						&initial_freq);
}

int rqbalance_pm_notify(struct notifier_block *notify_block,
				unsigned long mode, void *unused)
{
	switch (mode) {
	case PM_HIBERNATION_PREPARE:
	case PM_SUSPEND_PREPARE:
		rqbalance_state = IDLE;
		cancel_delayed_work_sync(&rqbalance_work);
		break;
	case PM_POST_SUSPEND:
	case PM_POST_HIBERNATION:
	case PM_POST_RESTORE:
		cpuquiet_wake_cpu(1, false);
		cpuquiet_wake_cpu(2, false);
		cpuquiet_wake_cpu(3, false);
		rqbalance_state = UP;
		rqbalance_kickstart();
		break;
	}

	return NOTIFY_OK;
}

/*
 * Retrieves frequency informations from CPUFreq.
 * This is needed for RQBalance's hotplugging decider
 * algorithms.
 *
 * \return Returns success (0) or negative errno.
 */
static int rqbalance_get_package_info(void)
{
	struct cpufreq_policy *policy;
	struct cpufreq_frequency_table *table;
	int count, i, prev_cluster = -1, cur_cluster;

	/* Paranoid initialization is needed in some conditions */
	num_of_cores[CLUSTER_LITTLE] = 0;
	num_of_cores[CLUSTER_BIG] = 0;

	for_each_possible_cpu(i) {
		cur_cluster = topology_physical_package_id(i);
		num_of_cores[cur_cluster]++;
		if (cur_cluster == prev_cluster)
			continue;

		prev_cluster = cur_cluster;
		available_clusters++;

		rqb_freq_limits[cur_cluster].main_cpu = i;
		rqb_freq_limits[cur_cluster].freq_min = 0;
		rqb_freq_limits[cur_cluster].freq_max = UINT_MAX;

		/*
		 * Get CPUFreq frequency table. RQBALANCE only works with
		 * tables composed of at least 4 frequency entries.
		 * This requirement has to be fullfilled for ALL clusters.
		 */
		policy = cpufreq_cpu_get(i);
		if (!policy)
			return -EINVAL;

		table = policy->freq_table;
		if (!table)
			return -EINVAL;

		for (count = 0;
		     table[count].frequency != CPUFREQ_TABLE_END;
		     count++);

		if (count < 4)
			return -EINVAL;

		idle_top_freq[cur_cluster] =
				table[(count / 2) - 1].frequency;
		idle_bottom_freq[cur_cluster] =
				table[(count / 2) - 2].frequency;
	};

	/* TODO: Check if we are effectively using HMP!!! This is NOT OK!!! */
	soc_is_hmp = (available_clusters > 1) ? true : false;
	pr_info("rqbalance: running as %s in %s mode\n",
		soc_is_hmp ? "hmp" : "smp",
		rqbalance_governor.use_isolation ? "isolation" : "hotplug");

	return 0;
}

static void rqbalance_stop(void)
{
	/*
	   first unregister the notifiers. This ensures the governor state
	   can't be modified by a cpufreq transition
	*/
	cpufreq_unregister_notifier(&balanced_cpufreq_nb,
		CPUFREQ_TRANSITION_NOTIFIER);

	cpufreq_unregister_notifier(&frequency_limits_nb,
		CPUFREQ_POLICY_NOTIFIER);

	cpuhp_remove_state_nocalls(CPUHP_AP_ONLINE_IDLE);

	unregister_pm_notifier(&pm_notifier_block);

	/* now we can force the governor to be idle */
	rqbalance_state = IDLE;
	cancel_delayed_work_sync(&rqbalance_work);
	destroy_workqueue(rqbalance_wq);
	del_timer(&load_timer);

	stop_rq_work();
	kfree(rq_data);

	rqbalance_sysfs_exit();
}

static int rqbalance_cfl_start(void)
{
	int rc = 0;

	cpufreq_register_notifier(&frequency_limits_nb,
		CPUFREQ_POLICY_NOTIFIER);
	
	rc = cpuhp_setup_state_nocalls(CPUHP_AP_ONLINE_IDLE,
		"rqbalance_cpuhp", cfl_hotplug_notify, NULL);
	if (rc < 0)
		goto end;
	rqb_hp_online = rc;
	rc = 0;
end:
	return rc;
}

static int rqbalance_start(void)
{
	int err, i, max_cpu_id = 0;

	err = rqbalance_sysfs_init();
	if (err)
		return err;

	avail_cpus_mask = rqbalance_governor.use_isolation ?
				cpu_unisolated_mask : cpu_online_mask;

	rqbalance_wq = alloc_workqueue("cpuquiet-rqbalance",
			WQ_UNBOUND | WQ_FREEZABLE, 1);
	if (!rqbalance_wq)
		return -ENOMEM;

	INIT_DELAYED_WORK(&rqbalance_work, rqbalance_work_func);

	up_delay = msecs_to_jiffies(UPCORE_DELAY_MS);
	down_delay = msecs_to_jiffies(DNCORE_DELAY_MS);
	recheck_delay = msecs_to_jiffies(RECHECK_DELAY_MS);

	err = rqbalance_get_package_info();
	if (err)
		return err;

	for_each_possible_cpu(i)
		max_cpu_id++;

	/* Set value as target MAX on-line number of CPUs */
	if (nr_run_thresholds[max_cpu_id] != UINT_MAX)
		nr_run_thresholds[max_cpu_id] = UINT_MAX;

	/* HACK: Adjust dual-core thresholds for non-HMP SoCs */
	if (!soc_is_hmp) {
		nr_run_thresholds[0] = 250;
		nr_down_run_thresholds[1] = 110;
	}

	cpufreq_register_notifier(&balanced_cpufreq_nb,
		CPUFREQ_TRANSITION_NOTIFIER);

	err = rqbalance_cfl_start();
	if (err)
		pr_warn("%s: Couldn't start RQB-CFL!\n", __func__);

	init_timer(&load_timer);
	load_timer.function = calculate_load_timer;

	rqbalance_kickstart();

	rq_data = kzalloc(sizeof(struct runqueue_data), GFP_KERNEL);
	if (rq_data == NULL) {
		pr_err("%s cannot allocate memory\n", __func__);
		return -ENOMEM;
	}
	spin_lock_init(&rq_data->lock);
	rq_data->nr_run_sample_head = (uint8_t)-1;
	rq_data->nr_run_sample_tail = (uint8_t)-1;
	rq_data->nr_run_sample_count = 0;
	rq_data->update_rate = (RQ_AVG_TIMER_RATE_NS / 1000000);
	INIT_DEFERRABLE_WORK(&rq_data->work, rq_work_fn);

	start_rq_work();

	pm_notifier_block.notifier_call = rqbalance_pm_notify;
	register_pm_notifier(&pm_notifier_block);

	return 0;
}

struct cpuquiet_governor rqbalance_governor = {
	.name		= "rqbalance",
	.start		= rqbalance_start,
	.stop		= rqbalance_stop,
	.use_isolation	= false,
	.isol_switch_supported = true,
	.owner		= THIS_MODULE,
};

static int __init init_rqbalance(void)
{
	return cpuquiet_register_governor(&rqbalance_governor);
}

static void __exit exit_rqbalance(void)
{
	cpuquiet_unregister_governor(&rqbalance_governor);
}

MODULE_LICENSE("GPL");
#ifdef CONFIG_CPU_QUIET_DEFAULT_GOV_RQBALANCE
fs_initcall(init_rqbalance);
#else
module_init(init_rqbalance);
#endif
module_exit(exit_rqbalance);
